/* 
 * httpsQueue.c
 * 
 * communication between the worker and other threads is done via
 * RW-locks and data is passed via a queue.
 *
 * Since this queue is a FIFO we could use a tail queue. But we don't
 * like the malloc() overhead, so therefore we use a static array.
 *
 * RCS $Id: httpsQueue.c,v 1.3 2006/12/07 10:33:20 root Exp root $
 */

#include "sg.h"
#include "ufdb.h"
#include "ufdblib.h"
#include "httpsQueue.h"

#include <unistd.h>
#include <string.h>
#include <pthread.h>

#define UFDB_HTTPS_QUEUE_SIZE	 128

static struct {
   int    portnumber;
   char   hostname[1028];
} q[UFDB_HTTPS_QUEUE_SIZE];

static int    ihead = 0;                    /* array index for the head */
static int    itail = 0;                    /* array index for the tail */
static int    n_queued = 0;
static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  empty = PTHREAD_COND_INITIALIZER;


#if 0
void initHttpsQueue( void )
{
   ihead = 0;
   itail = 0;
   n_queued = 0;
   pthread_mutex_init( &lock, NULL );
   pthread_cond_init( &empty, NULL );
}
#endif


/*
 * enqueue a hostname:portnumber for future https tunnel detection.
 */
int ufdbHttpsQueueRequest( char * hostname, int portnumber )
{
   int i;

   pthread_mutex_lock( &lock );

   /*
    * prevent duplicates: check if the request is already queued.
    */
   for (i = ihead; i != itail; i = (i+1) % UFDB_HTTPS_QUEUE_SIZE)
   {
      if (strcmp( hostname, q[i].hostname ) == 0  &&
          q[i].portnumber == portnumber)
      {
	 pthread_mutex_unlock( &lock );
         return 0;
      }
   }

   if (n_queued < UFDB_HTTPS_QUEUE_SIZE)
   {
      /*
       * insert the request at the tail
       */
      q[itail].portnumber = portnumber;
      strcpy( q[itail].hostname, hostname );

      itail = (itail + 1) % UFDB_HTTPS_QUEUE_SIZE;
      n_queued++;

      /*
       * signal anyone who is waiting
       */
      pthread_cond_signal( &empty );
   } 
   else    
   {
      /* queue is full; just drop the request */
      ufdbLogError( "ufdbHttpsQueueRequest: https tunnel detection queue is full." );
      pthread_mutex_unlock( &lock );
      return -1;
   }

   /*
    * leave the critical section
    */
   pthread_mutex_unlock( &lock );

   return 0;
}


/*
 * Get a queued https tunnel check request.
 */
void ufdbGetHttpsRequest( char * hostname, int * portnumber )
{
   pthread_mutex_lock( &lock );

   while (1)
   {
      /*
       * if there are requests in the queue
       */
      if (n_queued > 0)
      {
         /*
          * get the request at the head
          */
         *portnumber = q[ihead].portnumber;
	 strcpy( hostname, q[ihead].hostname );
         ihead = (ihead + 1) % UFDB_HTTPS_QUEUE_SIZE;
         n_queued--;

         pthread_mutex_unlock( &lock );

	 return;
      } 
      else   /* otherwise wait until there are fds available */ 
      {
         pthread_cond_wait( &empty, &lock );
         /*
          * when I'm here, I've been signaled because the queue
          * is not empty.  Go try again.
          */
      }
   }
}

